#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class NeutronOpsV1(object):

    def __init__(self, handler):
        self.openstack_driver = handler.openstack_driver
        self.plugin = self.openstack_driver.plugin

    def hm_binding_count(self, context, hm_id):
        return self.openstack_driver._hm_binding_count(context, hm_id)

    def hm_get(self, context, hm_id):
        return self.openstack_driver._pool_get_hm(context, hm_id)

    def member_get_ip(self, context, member, use_float=False):
        return self.openstack_driver._member_get_ip(context, member, use_float)

    def member_count(self, context, member):
        return self.openstack_driver._member_count(context, member)

    def member_get(self, context, member_id):
        return self.openstack_driver._member_get(context, member_id)

    def pool_get(self, context, pool_id):
        return self.plugin.get_pool(context, pool_id)

    def pool_get_tenant_id(self, context, pool_id):
        return self.openstack_driver._pool_get_tenant_id(context, pool_id)

    def vip_get(self, context, vip_id):
        return self.plugin.get_vip(context, vip_id)

    def vip_get_id(self, context, pool_id):
        return self.openstack_driver._pool_get_vip_id(context, pool_id)



# Copyright 2014, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import acos_client

from a10_neutron_lbaas import a10_exceptions as ex
from a10_neutron_lbaas.db import models

import base


# The default set of plumbing hooks/scheduler, meant for hardware or manual orchestration

class PlumbingHooks(base.BasePlumbingHooks):

    def __init__(self, driver, devices=None, get_devices_func=None, **kwargs):
        super(PlumbingHooks, self).__init__(
            driver, devices=devices, get_devices_func=get_devices_func, **kwargs)
        if devices is not None:
            self.devices = devices
        elif get_devices_func is not None:
            self.devices = get_devices_func()
        else:
            self.devices = None
        self.appliance_hash = None

    def _late_init(self):
        if self.devices is None:
            self.devices = self.driver.config.get_devices()
        if self.appliance_hash is None:
            self.appliance_hash = acos_client.Hash(self.devices.keys())

    def _select_device_hash(self, tenant_id):
        self._late_init()

        # Must return device dict from config.py
        s = self.appliance_hash.get_server(tenant_id)
        return self.devices[s]

    def _select_device_db(self, tenant_id, db_session=None):
        self._late_init()

        # See if we have a saved tenant
        a10 = models.A10TenantBinding.find_by_tenant_id(tenant_id, db_session=db_session)
        if a10 is not None:
            if a10.device_name in self.devices:
                return self.devices[a10.device_name]
            else:
                raise ex.DeviceConfigMissing(
                    'A10 device %s mapped to tenant %s is not present in config; '
                    'add it back to config or migrate loadbalancers' %
                    (a10.device_name, tenant_id))

        # Nope, so we hash and save
        d = self._select_device_hash(tenant_id)
        models.A10TenantBinding.create_and_save(
            tenant_id=tenant_id, device_name=d['name'],
            db_session=db_session)

        return d

    def select_device(self, tenant_id, **kwargs):
        if self.driver.config.get('use_database'):
            return self._select_device_db(tenant_id)
        else:
            return self._select_device_hash(tenant_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""scaling_group_member_virtual_server_

Revision ID: 556f41dfda67
Revises: 2782f2bd94aa
Create Date: 2016-05-19 18:04:48.965907

"""

# revision identifiers, used by Alembic.
revision = '556f41dfda67'
down_revision = '2782f2bd94aa'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        "a10_scaling_group_member_virtual_servers",
        sa.Column('id',
                  sa.String(36),
                  primary_key=True,
                  nullable=False),
        sa.Column('created_at', sa.DateTime, nullable=False),
        sa.Column('updated_at', sa.DateTime, nullable=False),
        sa.Column('member_id',
                  sa.String(36),
                  sa.ForeignKey(u'a10_scaling_group_members.id'),
                  nullable=False),
        sa.Column('neutron_id',
                  sa.String(36),
                  nullable=False),
        sa.Column('ip_address',
                  sa.String(50),
                  nullable=False),
        sa.Column('interface_ip_address',
                  sa.String(50),
                  nullable=True),
        sa.Column('sflow_uuid',
                  sa.String(36),
                  nullable=False),
    )


def downgrade():
    op.drop_table("a10_scaling_group_member_virtual_servers")



# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.from neutron.db import model_base

from oslo_log import log as logging

import a10_neutron_lbaas.neutron_ext.common.constants as constants
import a10_neutron_lbaas.neutron_ext.db.a10_device_instance as a10_device_instance

LOG = logging.getLogger(__name__)


class A10DeviceInstancePlugin(a10_device_instance.A10DeviceInstanceDbMixin):

    supported_extension_aliases = [constants.A10_DEVICE_INSTANCE_EXT]

    def get_a10_device_instances(self, context, filters=None, fields=None):
        LOG.debug(
            "A10DeviceInstancePlugin.get_a10_instances(): filters=%s, fields=%s",
            filters,
            fields)
        return super(A10DeviceInstancePlugin, self).get_a10_device_instances(
            context, filters=filters, fields=fields)

    def create_a10_device_instance(self, context, a10_device_instance):
        LOG.debug("A10DeviceInstancePlugin.create(): a10_device_instance=%s", a10_device_instance)
        return super(A10DeviceInstancePlugin, self).create_a10_device_instance(context,
                                                                               a10_device_instance)

    def get_a10_device_instance(self, context, id, fields=None):
        LOG.debug("A10DeviceInstancePlugin.get_a10_instance(): id=%s, fields=%s",
                  id, fields)
        return super(A10DeviceInstancePlugin, self).get_a10_device_instance(context,
                                                                            id,
                                                                            fields=fields)

    # def update_a10_device_instance(self, context, id, a10_device_instance):
    #     LOG.debug(
    #         "A10DeviceInstancePlugin.update_a10_device_instance(): id=%s, instance=%s",
    #         id,
    #         a10_device_instance)
    #     return super(A10DeviceInstancePlugin, self).a10_device_instance(
    #         context,
    #         id,
    #         a10_device_instance)

    # def delete_a10_device_instance(self, context, id):
    #     LOG.debug("A10DeviceInstancePlugin.delete(): id=%s", id)
    #     return super(A10DeviceInstancePlugin, self).delete_a10_device_instance(context, id)



# Copyright 2014, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import acos_client

from a10_neutron_lbaas import a10_exceptions as ex


class BasePlumbingHooks(object):

    def __init__(self, driver, **kwargs):
        self.driver = driver
        self.client_wrapper_class = None

    # While you can override select_device in hooks to get custom selection
    # behavior, it is much easier to use the 'device_scheduling_filters'
    # mechanism, as documented in the config file.

    def select_device(self, tenant_id, **kwargs):
        # Not a terribly useful scheduler
        raise ex.NotImplemented()

    # Newer scheduling interface, which is called in favor the one above,
    # if present.

    # def select_device_with_lbaas_obj(self, tenant_id, a10_context, lbaas_obj,
    #                                  db_session=None, **kwargs):
    #     raise ex.NotImplemented()

    def get_a10_client(self, device_info, **kwargs):
        return acos_client.Client(
            device_info['host'], device_info['api_version'],
            device_info['username'], device_info['password'],
            port=device_info['port'], protocol=device_info['protocol'])

    # Network plumbing hooks from here on out

    def partition_create(self, client, os_context, partition_name):
        client.system.partition.create(partition_name)

    def partition_delete(self, client, os_context, partition_name):
        client.system.partition.delete(partition_name)

    def after_member_create(self, a10_context, os_context, member):
        pass

    def after_member_update(self, a10_context, os_context, member):
        pass

    def after_member_delete(self, a10_context, os_context, member):
        pass

    def after_vip_create(self, a10_context, os_context, vip):
        pass

    def after_vip_update(self, a10_context, os_context, vip):
        pass

    def after_vip_delete(self, a10_context, os_context, vip):
        pass

    def a10_context_exit_final(self, a10_context):
        pass

    def after_select_partition(self, a10_context):
        pass



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

import a10_neutron_lbaas.tests.test_case as test_case

import a10_neutron_lbaas.a10_openstack_lb as a10_openstack_lb
import a10_neutron_lbaas.plumbing_hooks as plumbing_hooks


class SetupA10OpenstackLBBase(object):

    @property
    def a10_openstack_lb_class(self):
        return a10_openstack_lb.A10OpenstackLBBase

    @property
    def a10_openstack_lb_kws(self):
        return {}


class SetupA10OpenstackLBV1(object):
    @property
    def a10_openstack_lb_class(self):
        return a10_openstack_lb.A10OpenstackLBV1


class SetupA10OpenstackLBV2(object):
    @property
    def a10_openstack_lb_class(self):
        return a10_openstack_lb.A10OpenstackLBV2


class SetupPlumbingHooks(object):
    @property
    def a10_openstack_lb_kws(self):
        kws = super(SetupPlumbingHooks, self).a10_openstack_lb_kws
        kws2 = kws.copy()
        kws2['plumbing_hooks_class'] = plumbing_hooks.PlumbingHooks
        return kws2


class TestA10Openstack(SetupA10OpenstackLBBase, test_case.TestCase):

    def setUp(self):
        mock_driver = mock.MagicMock()
        with mock.patch('acos_client.Client'):
            self.a = self.a10_openstack_lb_class(mock_driver, **self.a10_openstack_lb_kws)

    def test_sanity(self):
        pass


class TestA10OpenstackPlumbingHooks(SetupPlumbingHooks, TestA10Openstack):
    pass


class TestA10OpenstackV1(SetupA10OpenstackLBV1, TestA10Openstack):
    pass


class TestA10OpenstackV1PlumbingHooks(SetupA10OpenstackLBV1, SetupPlumbingHooks, TestA10Openstack):
    pass


class TestA10OpenstackV2(SetupA10OpenstackLBV2, TestA10Openstack):
    pass


class TestA10OpenstackV2PlumbingHooks(SetupA10OpenstackLBV2, SetupPlumbingHooks, TestA10Openstack):
    pass



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import test_base

import a10_neutron_lbaas.a10_exceptions as a10_ex


class TestLB(test_base.UnitTestBase):

    def test_create(self):
        m = test_base.FakeLoadBalancer()
        self.a.lb.create(None, m)
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('call.slb.virtual_server.create' in s)
        self.assertTrue('fake-lb-id-001' in s)
        self.assertTrue('5.5.5.5' in s)
        self.assertTrue('UP' in s)

    def test_create_default_vrid_none_v21(self):
        self._test_create_default_vrid("2.1", None)

    def test_create_default_vrid_set_v21(self):
        self._test_create_default_vrid("2.1", 7)

    def test_create_default_vrid_none_v30(self):
        self._test_create_default_vrid("3.0", None)

    def test_create_default_vrid_set_v30(self):
        self._test_create_default_vrid("3.0", 7)

    def _test_create_default_vrid(self, api_ver=None, default_vrid=None):

        """
        Due to how the config is pulled in, we override the config
        for all of the devices.
        """

        for k, v in self.a.config.get_devices().items():
            v['api_version'] = api_ver
            v['default_virtual_server_vrid'] = default_vrid

        lb = test_base.FakeLoadBalancer()
        self.a.lb.create(None, lb)

        create = self.a.last_client.slb.virtual_server.create
        create.assert_has_calls([mock.ANY])
        calls = create.call_args_list

        if default_vrid is not None:
            self.assertIn('vrid=%s' % default_vrid, str(calls))
        if default_vrid is None:
            foundVrid = any(
                'vrid' in x.get('axapi_args', {}).get('virtual_server', {})
                for (_, x) in calls)
            self.assertFalse(
                foundVrid,
                'Expected to find no vrid in {0}'.format(str(calls)))

    # There's no code that causes listeners to be added
    # if they are present when the pool is created.
    # We'd use unittest.skip if it worked with cursed 2.6

    # def test_create_with_listeners(self):
    #     pool = test_base.FakePool('HTTP', 'ROUND_ROBIN', None)
    #     m = test_base.FakeLoadBalancer()
    #     for x in [1, 2, 3]:
    #         z = test_base.FakeListener('TCP', 2222+x, pool=pool,
    #                                    loadbalancer=m)
    #         m.listeners.append(z)
    #     self.a.lb.create(None, m)
    #     s = str(self.a.last_client.mock_calls)
    #     print ("LAST CALLS {0}".format(s))
    #     self.assertTrue('call.slb.virtual_server.create' in s)
    #     self.assertTrue('fake-lb-id-001' in s)
    #     self.assertTrue('5.5.5.5' in s)
    #     self.assertTrue('UP' in s)
    #     self.assertTrue('vport.create' in s)
    #     for x in [1, 2, 3]:
    #         self.assertTrue(str(2222+x) in s)

    def test_update_down(self):
        m = test_base.FakeLoadBalancer()
        m.admin_state_up = False
        self.a.lb.update(None, m, m)
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('call.slb.virtual_server.update' in s)
        self.assertTrue('fake-lb-id-001' in s)
        self.assertTrue('5.5.5.5' in s)
        self.assertTrue('DOWN' in s)

    def test_delete(self):
        m = test_base.FakeLoadBalancer()
        self.a.lb.delete(None, m)
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('call.slb.virtual_server.delete' in s)
        self.assertTrue(m.id in s)

    def test_delete_removes_slb(self):
        m = test_base.FakeLoadBalancer()
        self.a.lb.delete(None, m)

    def test_refresh(self):
        try:
            self.a.lb.refresh(None, test_base.FakeLoadBalancer())
        except a10_ex.UnsupportedFeature:
            pass

    def test_stats(self):
        test_lb = test_base.FakeLoadBalancer()
        self.a.lb.stats(None, test_lb)

        self.print_mocks()

        s = str(self.a.last_client.mock_calls)
        self.assertTrue('call.slb.virtual_server.stats' in s)

    def do_raise_exception(self, e, msg="mock raised exception"):
        def raise_exception(e, msg="acos broke!"):
            raise e(msg)

        return lambda *args, **kwargs: raise_exception(e, msg)



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import a10_neutron_lbaas.tests.test_case as test_case
import mock

import a10_neutron_lbaas.a10_openstack_lb as a10_os
import a10_neutron_lbaas.plumbing_hooks as hooks


def _build_openstack_context():
    admin_context = {
        "tenant_id": "admin"
    }

    return mock.Mock(admin_context=admin_context)


class FakeA10OpenstackLB(object):

    def __init__(self, openstack_driver, **kw):
        super(FakeA10OpenstackLB, self).__init__(
            mock.MagicMock(),
            **kw)
        self.openstack_driver = mock.MagicMock()
        self.plumbing_hooks = hooks.PlumbingHooks(self)
        self.openstack_context = _build_openstack_context()

    def _get_a10_client(self, device_info, **kwargs):
        self.device_info = device_info
        self.last_client = mock.MagicMock()
        return self.last_client

    def reset_mocks(self):
        self.openstack_driver = mock.MagicMock()
        self.plumbing_hooks = hooks.PlumbingHooks(self)
        self.last_client = self._get_a10_client(self.device_info)
        return self.last_client


class FakeA10OpenstackLBV1(FakeA10OpenstackLB, a10_os.A10OpenstackLBV1):
    pass


class FakeA10OpenstackLBV2(FakeA10OpenstackLB, a10_os.A10OpenstackLBV2):

    def __init__(self, openstack_driver, **kw):
        super(FakeA10OpenstackLBV2, self).__init__(
            openstack_driver,
            neutron_hooks_module=mock.MagicMock(),
            **kw)
        self.certmgr = mock.Mock()


class UnitTestBase(test_case.TestCase):

    def _build_openstack_context(self):
        return _build_openstack_context()

    def setUp(self, openstack_lb_args={}):
        unit_dir = os.path.dirname(__file__)
        unit_config = os.path.join(unit_dir, "unit_config")
        os.environ['A10_CONFIG_DIR'] = unit_config

        if not hasattr(self, 'version') or self.version == 'v2':
            self.a = FakeA10OpenstackLBV2(mock.MagicMock(), **openstack_lb_args)
        else:
            self.a = FakeA10OpenstackLBV1(mock.MagicMock(), **openstack_lb_args)

    def print_mocks(self):
        print("OPENSTACK ", self.a.openstack_driver.mock_calls)
        print("CLIENT ", self.a.last_client.mock_calls)

    def empty_mocks(self):
        self.print_mocks()
        self.assertEqual(0, len(self.a.openstack_driver.mock_calls))
        self.assertEqual(0, len(self.a.last_client.mock_calls))

    def empty_close_mocks(self):
        self.print_mocks()
        self.assertEqual(0, len(self.a.openstack_driver.mock_calls))
        self.assertEqual(1, len(self.a.last_client.mock_calls))
        self.a.last_client.session.close.assert_called_with()



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import acos_client.errors as acos_errors

import handler_base_v2
import v2_context as a10

LOG = logging.getLogger(__name__)


class LoadbalancerHandler(handler_base_v2.HandlerBaseV2):

    def _set(self, set_method, c, context, lb):
        status = c.client.slb.UP
        if not lb.admin_state_up:
            status = c.client.slb.DOWN

        try:
            vip_meta = self.meta(lb, 'virtual_server', {})

            set_method(
                self._meta_name(lb),
                lb.vip_address,
                status,
                vrid=c.device_cfg.get('default_virtual_server_vrid'),
                axapi_body=vip_meta)
        except acos_errors.Exists:
            pass

    def _create(self, c, context, lb):
        self._set(c.client.slb.virtual_server.create, c, context, lb)

    def create(self, context, lb):
        with a10.A10WriteStatusContext(self, context, lb, action='create') as c:
            self._create(c, context, lb)
            self.hooks.after_vip_create(c, context, lb)

    def update(self, context, old_lb, lb):
        with a10.A10WriteStatusContext(self, context, lb) as c:
            self._set(c.client.slb.virtual_server.update, c, context, lb)
            self.hooks.after_vip_update(c, context, lb)

    def _delete(self, c, context, lb):
        try:
            c.client.slb.virtual_server.delete(self._meta_name(lb))
        except acos_errors.NotFound:
            pass

    def delete(self, context, lb):
        with a10.A10DeleteContext(self, context, lb) as c:
            self._delete(c, context, lb)
            self.hooks.after_vip_delete(c, context, lb)

    def stats(self, context, lb):
        with a10.A10Context(self, context, lb) as c:
            try:
                name = self.meta(lb, 'id', lb.id)
                r = c.client.slb.virtual_server.stats(name)

                return {
                    "bytes_in": r["virtual_server_stat"]["req_bytes"],
                    "bytes_out": r["virtual_server_stat"]["resp_bytes"],
                    "active_connections":
                        r["virtual_server_stat"]["cur_conns"],
                    "total_connections": r["virtual_server_stat"]["tot_conns"]
                }
            except Exception:
                return {
                    "bytes_in": 0,
                    "bytes_out": 0,
                    "active_connections": 0,
                    "total_connections": 0
                }

    def refresh(self, context, lb):
        LOG.debug("LB Refresh called.")
        # Ensure all elements associated with this LB exist on the device.



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import test_base

from a10_neutron_lbaas import plumbing_hooks


class TestPlumbingHooks(test_base.UnitTestBase):

    def test_select_device(self):
        hooks = plumbing_hooks.PlumbingHooks(self.a)

        a = hooks.select_device("first-token")
        hooks.select_device("second-token")
        self.assertEqual(a, hooks.select_device("first-token"))



# Copyright 2016, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from acos_client import errors as acos_errors

LOG = logging.getLogger(__name__)


def initialize_licensing(vth_cfg, device_cfg, client):

    licensing = vth_cfg.get('license_manager')
    if licensing is not None:
        client.license_manager.paygo(
            llp_hosts=map(lambda host: host['ip'], licensing["hosts"]),
            sn=licensing["serial"],
            instance_name=device_cfg.get('name'),
            use_mgmt_port=licensing["use-mgmt-port"],
            bandwidth_base=licensing["bandwidth-base"],
            interval=licensing["interval"]
        )


def initialize_interfaces(vth_cfg, device_cfg, client):

    networks = vth_cfg.get("vthunder_data_networks", [])
    for x in range(1, len(networks) + 1):
        # Statically coded until we have devices that correctly configure
        # data-plane interfaces from dhcp
        client.interface.ethernet.update(x, dhcp=True, enable=True)


def initialize_dns(vth_cfg, device_cfg, client):
    dns = vth_cfg.get('dns_resolver')

    if dns is not None:
        client.dns.set(**dns)


def initialize_sflow(vth_cfg, device_cfg, client):

    collector = vth_cfg.get('sflow_collector')
    if collector is None:
        return

    try:
        client.sflow.collector.ip.create(collector['host'], collector['port'])
    except acos_errors.Exists:
        pass

    try:
        client.sflow.polling.create(http_counter=1)
    except acos_errors.Exists:
        pass

    try:
        client.sflow.setting.create(None, None, None, 1)
    except acos_errors.Exists:
        pass


def initialize_vthunder(a10_cfg, device_cfg, client):
    """Perform initialization of system-wide settings"""

    vth = a10_cfg.get_vthunder_config()
    initialize_interfaces(vth, device_cfg, client)
    initialize_dns(vth, device_cfg, client)
    initialize_licensing(vth, device_cfg, client)
    initialize_sflow(vth, device_cfg, client)



# Copyright 2015 A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import mock
import test_base

import a10_neutron_lbaas.a10_exceptions as a10_ex
from a10_neutron_lbaas import constants


LOG = logging.getLogger(__name__)


class TestListenersTerminatedHTTPS(test_base.UnitTestBase):

    def test_update_no_lb(self):
        m = test_base.FakeListener('TCP', 2222, pool=mock.MagicMock(),
                                   loadbalancer=None)
        try:
            self.a.listener.update(None, m, m)
        except a10_ex.UnsupportedFeature:
            pass

    def test_update_no_pool(self):
        m = test_base.FakeListener('HTTP', 8080, pool=None,
                                   loadbalancer=test_base.FakeLoadBalancer())
        self.a.listener.create(None, m)
        self.assertFalse('update' in str(self.a.last_client.mock_calls))

    def test_update(self):
        pool = test_base.FakePool(constants.PROTOCOL_TERMINATED_HTTPS,
                                  constants.LB_METHOD_ROUND_ROBIN, None)
        lb = test_base.FakeLoadBalancer()
        m = test_base.FakeListener(constants.PROTOCOL_TERMINATED_HTTPS, 2222,
                                   pool=pool, loadbalancer=lb)
        certmgr = FakeCertManager()

        self.a.barbican_client = certmgr
        # self.a.listener.set_certmgr(certmgr)

        pool.listener = m

        self.a.listener.update(None, m, m)

        self.print_mocks()
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('vport.update' in s)
        self.assertTrue('fake-lb-id-001' in s)
        self.assertTrue('fake-listen-id-001' in s)
        self.assertTrue('port=2222' in s)
        self.assertTrue('HTTPS' in s)

    def test_delete(self):
        pool = test_base.FakePool(constants.PROTOCOL_TERMINATED_HTTPS,
                                  constants.LB_METHOD_ROUND_ROBIN, None)
        lb = test_base.FakeLoadBalancer()
        m = test_base.FakeListener(constants.PROTOCOL_TERMINATED_HTTPS, 2222,
                                   pool=pool, loadbalancer=lb)

        pool.listener = m

        self.a.listener.delete(None, m)

        self.print_mocks()
        s = str(self.a.last_client.mock_calls)
        LOG.debug("DELETE RESULT %s" % s)
        self.assertTrue('vport.delete' in s)
        self.assertTrue('fake-lb-id-001' in s)
        self.assertTrue('fake-listen-id-001' in s)
        self.assertTrue('port=2222' in s)
        self.assertTrue('HTTPS' in s)

    def test_create_protocol_terminated_https(self):
        pool = test_base.FakePool(constants.PROTOCOL_TERMINATED_HTTPS,
                                  constants.LB_METHOD_ROUND_ROBIN, None)
        lb = test_base.FakeLoadBalancer()
        m = test_base.FakeListener(constants.PROTOCOL_TERMINATED_HTTPS, 2222,
                                   pool=pool, loadbalancer=lb)
        m.default_tls_container_id = "CONTAINERID"
        pool.listener = m

        certmgr = FakeCertManager()

        self.a.barbican_client = certmgr
        self.a.listener.create(None, m)
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('HTTPS' in s)

    @mock.patch('a10_neutron_lbaas.v2.handler_listener.certwrapper')
    def test_barbican_client_not_null(self, certmgr):
        """This tests that the barbican_client dependency is always there."""
        handler = self.a.listener
        handler.barbican_client = None
        pool = test_base.FakePool(constants.PROTOCOL_TERMINATED_HTTPS,
                                  constants.LB_METHOD_ROUND_ROBIN, None)
        lb = test_base.FakeLoadBalancer()
        m = test_base.FakeListener(constants.PROTOCOL_TERMINATED_HTTPS, 2222,
                                   pool=pool, loadbalancer=lb)
        pool.listener = m
        handler.create(None, m)
        self.assertTrue(handler.barbican_client is not None)

    def _tls_container_shared(self):
        certmgr = FakeCertManager()
        certmgr.certificate = "CERTDATA"
        certmgr.private_key = "PRIVATEKEY"
        certmgr.private_key_passphrase = "SECRETPASSWORD"

        handler = self.a.listener
        handler.barbican_client = certmgr
        pool = test_base.FakePool(constants.PROTOCOL_TERMINATED_HTTPS,
                                  constants.LB_METHOD_ROUND_ROBIN, None)
        lb = test_base.FakeLoadBalancer()
        m = test_base.FakeListener(constants.PROTOCOL_TERMINATED_HTTPS, 2222,
                                   pool=pool, loadbalancer=lb)
        return (certmgr, m, pool, handler)

    def test_create_tls_container_positive(self):
        """Test that cert data is passed to handler when data is available"""
        certmgr, listener, pool, handler = self._tls_container_shared()
        listener.default_tls_container_id = "CONTAINERID"

        pool.listener = listener
        handler.create(None, listener)
        s = str(self.a.last_client.mock_calls)
        self.assertTrue(certmgr.certificate in s)
        self.assertTrue(certmgr.private_key in s)
        self.assertTrue(certmgr.private_key_passphrase in s)

    def test_create_tls_container_negative(self):
        """Test that cert data is not passed to handler"""

        certmgr, listener, pool, handler = self._tls_container_shared()
        listener.default_tls_container_id = None

        pool.listener = listener
        handler.create(None, listener)
        s = str(self.a.last_client.mock_calls)
        self.assertFalse("cert=" in s)
        self.assertFalse("key=" in s)


class FakeCertManager(object):
    def __init__(self):
        self.get_private_key_value = ""
        self.get_certificate_value = ""
        self.get_private_key_passphrase_value = ""
        self.container_name = "tls-container"
        self.set_mocks()

    def set_mocks(self):
        self.mock_cert = mock.Mock(return_value=self.certificate)
        self.mock_key = mock.Mock(return_value=self.private_key)
        self.mock_passphrase = mock.Mock(return_value=self.private_key_passphrase)
        self.mock_cert_container = mock.Mock()
        self.mock_cert_container.configure_mock(name=self.container_name)

        self.mock_certificate_result = mock.Mock(return_value=mock.Mock(
                                                 get_certificate=self.mock_cert,
                                                 get_private_key=self.mock_key,
                                                 get_private_key_passphrase=self.mock_passphrase,
                                                 _cert_container=self.mock_cert_container))

        self.get_certificate = self.mock_certificate_result

    @property
    def private_key(self):
        return self.get_private_key_value

    @private_key.setter
    def private_key(self, value):
        self.mock_key.return_value = value

    @property
    def certificate(self):
        return self.get_certificate_value

    @certificate.setter
    def certificate(self, value):
        self.mock_cert.return_value = value

    @property
    def private_key_passphrase(self):
        return self.get_private_key_passphrase_value

    @private_key_passphrase.setter
    def private_key_passphrase(self, value):
        self.mock_passphrase.return_value = value



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import a10_neutron_lbaas.tests.unit.test_base as test_base


class FakeModel(dict, object):
    def __getitem__(self, key, default=None):
        attr = getattr(self, key, default)
        return attr

    def get(self, key, default=None):
        return getattr(self, key, default)

    # def copy(self):
    #     import copy
    #     return copy.deepcopy(self)


class FakeHM(FakeModel):
    def __init__(self, id="hm01", name="hm01"):
        self.id = id
        self.name = name
        self.pools = []


class FakePool(FakeModel):
    def __init__(self, id="p01", name="p01"):
        self.id = id
        self.name = name


class UnitTestBase(test_base.UnitTestBase):

    def __init__(self, *args):
        super(UnitTestBase, self).__init__(*args)
        self.version = 'v1'



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""create a10 tenant table

Revision ID: 579f359e6e30
Revises: None
Create Date: 2016-03-17 21:44:11.257126

"""

# revision identifiers, used by Alembic.
revision = '579f359e6e30'
down_revision = None
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'a10_tenant_bindings',
        sa.Column('id', sa.String(36), primary_key=True),
        sa.Column('created_at', sa.DateTime),
        sa.Column('updated_at', sa.DateTime),
        sa.Column('tenant_id', sa.String(36), nullable=False),
        sa.Column('device_name', sa.String(1024), nullable=False)
    )


def downgrade():
    op.drop_table('a10_tenant_bindings')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import sqlalchemy as sa
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy import inspect
from sqlalchemy.orm import backref, relationship

from a10_neutron_lbaas.db import model_base as models


LOG = logging.getLogger(__name__)


class A10ScalingGroup(models.A10Base):
    """A10 Scaling Group - container of switch and workers"""

    __tablename__ = u'a10_scaling_groups'

    id = sa.Column(sa.String(36),
                   primary_key=True,
                   default=models._uuid_str,
                   nullable=False)
    tenant_id = sa.Column(sa.String(255), nullable=True)
    name = sa.Column(sa.String(255), nullable=True)
    description = sa.Column(sa.String(255), nullable=True)
    scaling_policy_id = sa.Column(sa.String(36),
                                  sa.ForeignKey('a10_scaling_policies.id'),
                                  nullable=True)

    scaling_policy = relationship('A10ScalingPolicy', backref='scaling_groups')
    switches = relationship('A10ScalingGroupSwitch')
    workers = relationship('A10ScalingGroupWorker')
    members = relationship('A10ScalingGroupMember', backref='scaling_group')

    __mapper_args__ = {
        'polymorphic_identity': __tablename__
    }


class A10ScalingGroupBinding(models.A10Base):
    __tablename__ = u'a10_scaling_group_bindings'

    id = sa.Column(sa.String(36),
                   primary_key=True,
                   nullable=False,
                   default=models._uuid_str)
    scaling_group_id = sa.Column(sa.String(36),
                                 sa.ForeignKey('a10_scaling_groups.id'),
                                 nullable=False)
    scaling_group = relationship(A10ScalingGroup, backref='bindings')

    lbaas_loadbalancer_id = sa.Column(sa.String(36),
                                      unique=True,
                                      nullable=False)


class A10ScalingGroupMember(models.A10Base):
    """A10 Scaling Group Member - switch/worker depending on 'role'"""
    __tablename__ = "a10_scaling_group_members"

    id = sa.Column(sa.String(36),
                   primary_key=True,
                   default=models._uuid_str,
                   nullable=False)
    type = sa.Column(sa.String(50), nullable=False)
    scaling_group_id = sa.Column(sa.String(36),
                                 sa.ForeignKey('a10_scaling_groups.id'),
                                 nullable=False)
    tenant_id = sa.Column(sa.String(255), nullable=True)
    name = sa.Column(sa.String(255), nullable=True)
    description = sa.Column(sa.String(255), nullable=True)
    host = sa.Column(sa.String(255), nullable=False)
    api_version = sa.Column(sa.String(12), nullable=False)
    username = sa.Column(sa.String(255), nullable=False)
    password = sa.Column(sa.String(255), nullable=False)
    protocol = sa.Column(sa.String(255), nullable=False)
    port = sa.Column(sa.Integer, nullable=False)
    nova_instance_id = sa.Column(sa.String(36), nullable=False)

    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
        'polymorphic_on': type
    }

    def add_virtual_server(self, neutron_id, **kwargs):
        vs = A10ScalingGroupMemberVirtualServer.create(
            neutron_id=neutron_id,
            **kwargs)
        self.virtual_servers.append(vs)
        return vs

    def get_virtual_server(self, neutron_id):
        return inspect(self).session.\
            query(A10ScalingGroupMemberVirtualServer).\
            filter_by(member_id=self.id, neutron_id=neutron_id).\
            one_or_none()

    def delete_virtual_server(self, neutron_id):
        vs = self.get_virtual_server(neutron_id)
        if vs:
            inspect(self).session.delete(vs)


class A10ScalingGroupWorker(A10ScalingGroupMember):
    __tablename__ = "a10_scaling_group_workers"

    id = sa.Column(sa.String(36),
                   sa.ForeignKey(u'a10_scaling_group_members.id'),
                   primary_key=True,
                   default=models._uuid_str,
                   nullable=False)

    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }


class A10ScalingGroupSwitch(A10ScalingGroupMember):
    __tablename__ = "a10_scaling_group_switches"

    id = sa.Column(sa.String(36),
                   sa.ForeignKey(u'a10_scaling_group_members.id'),
                   primary_key=True,
                   default=models._uuid_str,
                   nullable=False)

    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }


class A10ScalingGroupMemberVirtualServer(models.A10Base):
    __tablename__ = "a10_scaling_group_member_virtual_servers"

    id = sa.Column(sa.String(36),
                   primary_key=True,
                   default=models._uuid_str,
                   nullable=False)
    member_id = sa.Column(sa.String(36),
                          sa.ForeignKey(u'a10_scaling_group_members.id'),
                          nullable=False)
    member = relationship('A10ScalingGroupMember',
                          backref=backref('virtual_servers', cascade='all, delete-orphan'))
    neutron_id = sa.Column(sa.String(36),
                           nullable=False)

    ip_address = sa.Column(sa.String(50), nullable=False)
    interface_ip_address = sa.Column(sa.String(50), nullable=True)
    sflow_uuid = sa.Column(sa.String(36), nullable=False)

    def add_port(self, port, **kwargs):
        vs = A10ScalingGroupMemberVirtualServerPort.create(
            port=port,
            **kwargs)
        self.ports.append(vs)
        return vs

    def get_port(self, port):
        return inspect(self).session.\
            query(A10ScalingGroupMemberVirtualServerPort).\
            filter_by(virtual_server_id=self.id, port=port).\
            one_or_none()

    def delete_port(self, port):
        port = self.get_port(port)
        if port:
            inspect(self).session.delete(port)


class A10ScalingGroupMemberVirtualServerPort(models.A10Base):
    __tablename__ = "a10_scaling_group_member_virtual_server_ports"

    id = sa.Column(sa.String(36),
                   primary_key=True,
                   default=models._uuid_str,
                   nullable=False)
    virtual_server_id = sa.Column(sa.String(36),
                                  sa.ForeignKey(u'a10_scaling_group_member_virtual_servers.id'),
                                  nullable=False)
    virtual_server = relationship('A10ScalingGroupMemberVirtualServer',
                                  backref=backref('ports', cascade='all, delete-orphan'))
    port = sa.Column(sa.Integer,
                     nullable=False)
    protocol = sa.Column(sa.String(255), nullable=False)
    sflow_uuid = sa.Column(sa.String(36), nullable=False)


class A10ScalingPolicy(models.A10Base):
    __tablename__ = "a10_scaling_policies"

    id = sa.Column(sa.String(36),
                   primary_key=True,
                   default=models._uuid_str,
                   nullable=False)
    tenant_id = sa.Column(sa.String(255), nullable=True)
    name = sa.Column(sa.String(255), nullable=True)
    description = sa.Column(sa.String(255), nullable=True)

    cooldown = sa.Column(sa.Integer, nullable=False)
    min_instances = sa.Column(sa.Integer, nullable=False)
    max_instances = sa.Column(sa.Integer, nullable=True)

    reactions = relationship('A10ScalingPolicyReaction',
                             order_by="A10ScalingPolicyReaction.position",
                             collection_class=ordering_list('position'),
                             backref='policy')

    def scaling_group_ids(self):
        return [sg.id for sg in self.scaling_groups]


class A10ScalingPolicyReaction(models.A10Base):
    __tablename__ = "a10_scaling_policy_reactions"

    # A surrogate key is required by ordering_list
    id = sa.Column(sa.String(36),
                   primary_key=True,
                   default=models._uuid_str,
                   nullable=False)
    scaling_policy_id = sa.Column(sa.String(36),
                                  sa.ForeignKey('a10_scaling_policies.id'),
                                  nullable=False)
    position = sa.Column(sa.Integer,
                         nullable=False)
    alarm_id = sa.Column(sa.String(36),
                         sa.ForeignKey('a10_scaling_alarms.id'),
                         nullable=False)
    action_id = sa.Column(sa.String(36),
                          sa.ForeignKey('a10_scaling_actions.id'),
                          nullable=False)

    alarm = relationship('A10ScalingAlarm', backref='reactions')
    action = relationship('A10ScalingAction', backref='reactions')


class A10ScalingAlarm(models.A10Base):
    __tablename__ = "a10_scaling_alarms"

    id = sa.Column(sa.String(36),
                   primary_key=True,
                   default=models._uuid_str,
                   nullable=False)
    tenant_id = sa.Column(sa.String(255), nullable=True)
    name = sa.Column(sa.String(255), nullable=True)
    description = sa.Column(sa.String(255), nullable=True)

    aggregation = sa.Column(sa.String(50), nullable=False)
    measurement = sa.Column(sa.String(50), nullable=False)
    operator = sa.Column(sa.String(50), nullable=False)
    threshold = sa.Column(sa.Float(), nullable=False)
    unit = sa.Column(sa.String(50), nullable=False)
    period = sa.Column(sa.Integer, nullable=False)
    period_unit = sa.Column(sa.String(50), nullable=False)

    def scaling_group_ids(self):
        return set(x
                   for reaction in self.reactions
                   for x in reaction.policy.scaling_group_ids())


class A10ScalingAction(models.A10Base):
    __tablename__ = "a10_scaling_actions"

    id = sa.Column(sa.String(36),
                   primary_key=True,
                   default=models._uuid_str,
                   nullable=False)
    tenant_id = sa.Column(sa.String(255), nullable=True)
    name = sa.Column(sa.String(255), nullable=True)
    description = sa.Column(sa.String(255), nullable=True)

    action = sa.Column(sa.String(50), nullable=False)
    amount = sa.Column(sa.Integer)

    def scaling_group_ids(self):
        return set(x
                   for reaction in self.reactions
                   for x in reaction.policy.scaling_group_ids())



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import test_base


class TestHM(test_base.UnitTestBase):

    def assert_hm(self, mon_type, method, url, expect_code):
        self.print_mocks()
        self.a.last_client.slb.hm.create.assert_called_with(
            'abcdef', mon_type, '5', 5, '5',
            url=url, method=method, expect_code=expect_code,
            axapi_args={})

    def fake_hm(self, type):
        hm = {
            'tenant_id': 'tenv1',
            'id': 'abcdef',
            'name': 'abcdef',
            'type': type,
            'delay': '5',
            'timeout': 5,
            'max_retries': '5',
            'pools': [],
        }
        if type in ['HTTP', 'HTTPS']:
            hm['http_method'] = 'GET'
            hm['url_path'] = '/'
            hm['expected_codes'] = '200'
        return hm.copy()

    def test_create_ping(self):
        self.a.hm.create(None, self.fake_hm('PING'), 'p01')
        self.assert_hm(self.a.last_client.slb.hm.ICMP, None, None, None)
        pool_name = self.a.hm._pool_name(None, 'p01')
        self.a.last_client.slb.service_group.update.assert_called_with(
            pool_name, health_monitor='abcdef')

    def test_create_tcp(self):
        hm = self.fake_hm('TCP')
        hm['pools'] = [{'pool_id': 'p02'}, {'pool_id': 'p01'}]
        self.a.hm.create(None, hm, 'p01')
        self.assert_hm(self.a.last_client.slb.hm.TCP, None, None, None)
        pool_name = self.a.hm._pool_name(None, 'p02')
        self.a.last_client.slb.service_group.update.assert_called_with(
            pool_name, health_monitor='abcdef')

    def test_create_http(self):
        self.a.hm.create(None, self.fake_hm('HTTP'), 'p01')
        self.assert_hm(self.a.last_client.slb.hm.HTTP, 'GET', '/', '200')
        pool_name = self.a.hm._pool_name(None, 'p01')
        self.a.last_client.slb.service_group.update.assert_called_with(
            pool_name, health_monitor='abcdef')

    def test_create_https(self):
        self.a.hm.create(None, self.fake_hm('HTTPS'), 'p01')
        self.assert_hm(self.a.last_client.slb.hm.HTTPS, 'GET', '/', '200')
        pool_name = self.a.hm._pool_name(None, 'p01')
        self.a.last_client.slb.service_group.update.assert_called_with(
            pool_name, health_monitor='abcdef')

    def test_update_tcp(self, m_old=None, m=None):
        if m_old is None:
            m_old = self.fake_hm('TCP')
        if m is None:
            m = self.fake_hm('TCP')
        m['delay'] = 20
        self.a.hm.update(None, m_old, m, 'p01')
        self.a.last_client.slb.hm.update.assert_called_with(
            'abcdef',
            self.a.last_client.slb.hm.TCP, 20, 5, '5',
            url=None, method=None, expect_code=None,
            axapi_args={})

    def test_delete(self):
        expected = test_base.FakePool()
        fakehm = test_base.FakeHM()
        fakehm['tenant_id'] = 'tenv1'
        fakehm['id'] = 'fedcba'
        fakehm.pools.append(expected)

        self.a.hm.openstack_driver.plugin.get_pool.return_value = expected
        self.a.hm.openstack_driver._hm_binding_count.return_value = 1

        self.a.hm.delete(None, fakehm, 'p01')
        self.a.last_client.slb.hm.delete.assert_called_with(fakehm["id"])

    def test_delete_updates_pool_health_monitor(self):
        expected = test_base.FakePool()
        fakehm = test_base.FakeHM()
        fakehm['tenant_id'] = 'tenv1'
        fakehm['id'] = 'fedcba'
        fakehm['pools'] = []
        fakehm['pools'].append(expected)

        self.a.hm.openstack_driver._pool_get_hm.return_value = fakehm
        self.a.hm.openstack_driver.plugin.get_pool.return_value = expected
        self.a.hm.openstack_driver._hm_binding_count.return_value = 1

        pool_name = self.a.hm._pool_name(None, 'p01')
        self.a.hm.delete(None, fakehm, 'p01')

        self.a.last_client.slb.service_group.update.assert_called_with(
            pool_name, health_monitor='', health_check_disable=True)

    def test_dissociate_calls_service_group_update(self):
        fake_pool = test_base.FakePool()
        fake_hm = test_base.FakeHM()
        fake_hm["id"] = "id1"
        fake_hm["pools"] = []
        fake_hm["pools"].append(fake_pool)
        fake_hm['tenant_id'] = "tenv1"

        self.a.hm.dissociate(self.a.last_client, None, fake_hm, fake_pool.id)
        self.a.last_client.slb.service_group.update.assert_called(
            fake_pool.id, health_monitor="", health_check_disable=True)

    def test_dissociate_calls_hm_delete(self):
        fake_pool = test_base.FakePool()
        fake_hm = test_base.FakeHM()
        fake_pool["health_monitor_status"] = [{"monitor_id": fake_hm.id}]
        fake_hm["id"] = "id1"
        fake_hm["pools"] = []
        fake_hm["pools"].append(fake_pool)
        fake_hm['tenant_id'] = "tenv1"

        self.a.hm.dissociate(self.a.last_client, None, fake_hm, fake_pool.id)
        self.a.last_client.hm.service_group.delete.assert_called(self.a.hm._meta_name(fake_hm))



# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.from neutron.db import model_base

import mock

import a10_neutron_lbaas.tests.unit.test_a10_openstack_lb as test_a10_openstack_lb
import test_base

import a10_neutron_lbaas.a10_context as a10_context
import a10_neutron_lbaas.vthunder.instance_manager as instance_manager


class TestA10Context(test_a10_openstack_lb.SetupA10OpenstackLBBase, test_base.UnitTestBase):

    def setUp(self):
        super(TestA10Context, self).setUp()

        # unit_config.setUp()

        mock_driver = mock.MagicMock()
        with mock.patch('acos_client.Client'):
            a10_driver = self.a10_openstack_lb_class(
                mock_driver,
                **self.a10_openstack_lb_kws)

        mock_handler = mock.MagicMock(openstack_driver=mock_driver, a10_driver=a10_driver)

        session = self.open_session()
        mock_openstack_context = mock.MagicMock(session=session)

        mock_openstack_lbaas_obj = mock.MagicMock(tenant_id='fake-tenant-id')

        self.a10_context = a10_context.A10Context(mock_handler,
                                                  mock_openstack_context,
                                                  mock_openstack_lbaas_obj)

    def test_inventory_find(self):
        """Test inventory and scheduling hooks in their almost-real environment"""

        # We really just want to patch glance, but oslo library loading is non-standard.
        mock_glance_client = mock.MagicMock()
        mock_glance_client.images.list.return_value = []

        InstanceManager = instance_manager.InstanceManager
        with mock.patch(
            'a10_neutron_lbaas.vthunder.instance_manager.InstanceManager'
        ) as mock_instance_manager:

            mock_instance_manager.side_effect = lambda *args, **kwargs: InstanceManager(
                *args,
                glance_api=mock_glance_client,
                **kwargs)

            # with self.a10_context:
            #     device = self.a10_context._get_device()

        # session = self.a10_context.db_operations.session
        # session.commit()
        # saved_device = session.query(models.A10ApplianceSLB).get(appliance.id)

        # self.assertEqual(appliance, saved_appliance)


class TestA10ContextPlumbingHooks(test_a10_openstack_lb.SetupPlumbingHooks, TestA10Context):
    pass


class TestA10ContextV1(test_a10_openstack_lb.SetupA10OpenstackLBV1, TestA10Context):
    pass


class TestA10ContextV1PlumbingHooks(test_a10_openstack_lb.SetupA10OpenstackLBV1,
                                    test_a10_openstack_lb.SetupPlumbingHooks,
                                    TestA10Context):
    pass


class SetupA10ContextV2(test_a10_openstack_lb.SetupA10OpenstackLBV2):

    def setUp(self):
        super(SetupA10ContextV2, self).setUp()
        self.a10_context.openstack_lbaas_obj.root_loadbalancer = mock.MagicMock(id='fake-lb-id')


class TestA10ContextV2(SetupA10ContextV2, TestA10Context):
    pass


class TestA10ContextV2PlumbingHooks(SetupA10ContextV2,
                                    test_a10_openstack_lb.SetupPlumbingHooks,
                                    TestA10Context):
    pass



# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class FakeModel(object):
    def __init__(self, *args, **kwargs):
        for k, v in kwargs.iteritems():
            setattr(self, k, v)


class FakeInstance(dict):
    def __init__(self, **kwargs):
        _default_server = {
            "id": 42,
            "name": None,
            "image": None,
            "flavor": None,
            "meta": {},
            "files": {},
            "min_count": 1,  # optional extension
            "max_count": 1,  # optional extension
            "security_groups": [],
            "userdata": None,
            "key_name": None,  # optional extension
            "availability_zone": None,
            "block_device_mapping": None,  # optional extension
            "block_device_mapping_v2": None,  # optional extension
            "networks": ["mgmt-net"],  # optional extension
            "scheduler_hints": {},  # optional extension
            "config_drive": False,  # optional extension
            "disk_config": "AUTO",   # AUTO or MANUAL # optional extension
            "admin_pass": None  # optional extension
        }

        for k, v in _default_server.iteritems():
            self[k] = kwargs.get(k, _default_server.get(k, None))


class FakeImage(FakeModel):
    def __init__(self, id="image01", name="Image", metadata={}):
        super(FakeImage, self).__init__(id=id, name=name, metadata=metadata)


class FakeFlavor(FakeModel):
    def __init__(self, id="Flavor01", name="Flavor"):
        super(FakeFlavor, self).__init__(id=id, name=name)



# Copyright 2014-2016, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import logging

from a10_neutron_lbaas.acos import openstack_mappings
import acos_client.errors as acos_errors
import handler_base_v2
import handler_persist
import v2_context as a10

LOG = logging.getLogger(__name__)


class PoolHandler(handler_base_v2.HandlerBaseV2):

    def _set(self, set_method, c, context, pool, old_pool=None):
        self._update_session_persistence(old_pool, pool, c, context)
        args = {'service_group': self.meta(pool, 'service_group', {})}
        set_method(
            self._meta_name(pool),
            protocol=openstack_mappings.service_group_protocol(c, pool.protocol),
            lb_method=openstack_mappings.service_group_lb_method(c, pool.lb_algorithm),
            axapi_args=args)

        # session persistence might need a vport update
        if pool.listener:
            pool.listener.default_pool_id = pool.listener.default_pool_id or pool.id
            self.a10_driver.listener._update(c, context, pool.listener)

    def _create(self, c, context, pool):
        try:
            self._set(c.client.slb.service_group.create,
                      c, context, pool)
        except acos_errors.Exists:
            pass

    def create(self, context, pool):
        with a10.A10WriteStatusContext(self, context, pool) as c:
            try:
                self._set(c.client.slb.service_group.create,
                          c, context, pool)
            except acos_errors.Exists:
                pass

    def update(self, context, old_pool, pool):
        with a10.A10WriteStatusContext(self, context, pool) as c:

            self._set(c.client.slb.service_group.update,
                      c, context, pool, old_pool)

    def delete(self, context, pool):
        with a10.A10DeleteContext(self, context, pool) as c:
            debug_fmt = "handler_pool.delete(): removing member {0} from pool {1}"
            for member in pool.members:
                LOG.debug(debug_fmt.format(member, member.pool_id))
                self.a10_driver.member._delete(c, context, member)

            LOG.debug("handler_pool.delete(): Checking pool health monitor...")
            if pool.healthmonitor:
                # The pool.healthmonitor we get doesn't have a pool
                # Make a new one with the hm as the root
                hm = copy.copy(pool.healthmonitor)
                hm.pool = copy.copy(pool)
                hm.pool.healthmonitor = None
                LOG.debug("handler_pool.delete(): HM: %s" % hm)
                self.a10_driver.hm._delete(c, context, hm)

            try:
                c.client.slb.service_group.delete(self._meta_name(pool))
            except (acos_errors.NotFound, acos_errors.NoSuchServiceGroup):
                pass

            handler_persist.PersistHandler(
                c, context, pool, self._meta_name(pool)).delete()

    def _update_session_persistence(self, old_pool, pool, c, context):
        # didn't exist, does exist, create
        if not old_pool or (not old_pool.sessionpersistence and pool.sessionpersistence):
            p = handler_persist.PersistHandler(c, context, pool, old_pool)
            p.create()
            return

        # existed, change, delete and recreate
        if (old_pool.sessionpersistence and pool.sessionpersistence and
                old_pool.sessionpersistence.type != pool.sessionpersistence.type):
            p = handler_persist.PersistHandler(c, context, old_pool)
            p.delete()
            p = handler_persist.PersistHandler(c, context, pool)
            p.create()
            return

        # didn't exist, does exist now, create
        if old_pool.sessionpersistence and not pool.sessionpersistence:
            p = handler_persist.PersistHandler(c, context, pool)
            p.create()
            return

        # didn't exist, doesn't exist
        # did exist, does exist, didn't change
        return

    def stats(self, context, pool):
        result = {"stats": {}, "members": {}}
        with a10.A10Context(self, context, pool) as c:
            name = pool.id
            if name is not None:
                stats = c.client.slb.service_group.stats(name)
                result["stats"] = stats.get("stats", {})
                result["members"] = stats.get("members", {})

        return result



# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import six

from neutron.api import extensions
from neutron.api.v2 import attributes
from neutron.api.v2 import resource_helper
from neutron.common import exceptions
from neutron.services import service_base

import a10_neutron_lbaas.neutron_ext.common.constants as constants
from a10_neutron_lbaas.neutron_ext.common import resources
import a10_neutronclient.resources.a10_device_instance as a10_device_instance


RESOURCE_ATTRIBUTE_MAP = resources.apply_template(a10_device_instance.RESOURCE_ATTRIBUTE_MAP,
                                                  attributes)


# TODO(rename this to *Extension to avoid config file confusion)
class A10DeviceInstance(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "A10 Device Instances"

    @classmethod
    def get_alias(cls):
        return constants.A10_DEVICE_INSTANCE_EXT

    @classmethod
    def get_namespace(cls):
        return "http://docs.openstack.org/ext/neutron/a10_device_instance/api/v1.0"

    @classmethod
    def get_updated(cls):
        return "2015-11-18T16:17:00-07:00"

    @classmethod
    def get_description(cls):
        return ("A10 Device Instances")

    @classmethod
    def get_resources(cls):
        """Returns external resources."""
        my_plurals = resource_helper.build_plural_mappings(
            {}, RESOURCE_ATTRIBUTE_MAP)
        attributes.PLURALS.update(my_plurals)
        attr_map = RESOURCE_ATTRIBUTE_MAP
        resources = resource_helper.build_resource_info(my_plurals,
                                                        attr_map,
                                                        constants.A10_DEVICE_INSTANCE)

        return resources

    def update_attributes_map(self, attributes):
        super(A10DeviceInstance, self).update_attributes_map(
            attributes,
            extension_attrs_map=RESOURCE_ATTRIBUTE_MAP)

    def get_extended_resources(self, version):
        if version == "2.0":
            return RESOURCE_ATTRIBUTE_MAP
        else:
            return {}


class A10DeviceInstanceNotFoundError(exceptions.NotFound):
    def __init__(self, a10_device_instance_id):
        self.msg = _("A10 Device Instance {} could not be found.")
        super(A10DeviceInstanceNotFoundError, self).__init__()


class A10DeviceInstanceInUseError(exceptions.InUse):
    def __init__(self, a10_device_instance_id):
        self.message = _("A10 Device Instance is in use and cannot be deleted.")
        self.msg = self.message
        super(A10DeviceInstanceInUseError, self).__init__()


@six.add_metaclass(abc.ABCMeta)
class A10DeviceInstancePluginBase(service_base.ServicePluginBase):

    def get_plugin_name(self):
        return constants.A10_DEVICE_INSTANCE

    def get_plugin_description(self):
        return constants.A10_DEVICE_INSTANCE

    def get_plugin_type(self):
        return constants.A10_DEVICE_INSTANCE

    def __init__(self):
        super(A10DeviceInstancePluginBase, self).__init__()

    @abc.abstractmethod
    def get_a10_device_instances(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def create_a10_device_instance(self, context, device_instance):
        pass

    @abc.abstractmethod
    def get_a10_device_instance(self, context, id, fields=None):
        pass



# Copyright 2016,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from a10_neutron_lbaas.tests import test_case

from a10_neutron_lbaas.neutron_ext.common import resources


class TestResources(test_case.TestCase):

    def test_apply_template(self):
        template = {
            'none': None,
            'integer': 1,
            'string': 'a string',
            'tuple': ('a tuple', lambda x: x + 1, lambda x: x + 2),
            'dict': {'key': lambda x: x + 3},
            'list': ['a list', lambda x: x + 4]
        }

        actual = resources.apply_template(template, 0)

        expected = {
            'none': None,
            'integer': 1,
            'string': 'a string',
            'tuple': ('a tuple', 1, 2),
            'dict': {'key': 3},
            'list': ['a list', 4]
        }

        self.assertEqual(expected, actual)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy as sa

from a10_neutron_lbaas.db import model_base


class A10SLB(model_base.A10BaseMixin, model_base.A10Base):
    __tablename__ = 'a10_slbs'

    # For vip specific binding (as opposed to tenant level binding), this will
    # differ from A10TenantBinding, if that row exists at all.
    device_name = sa.Column(sa.String(1024), nullable=False)

    # LBaaS v1 or v2, only one of these will be defined
    pool_id = sa.Column(sa.String(36))

    loadbalancer_id = sa.Column(sa.String(36))

    @classmethod
    def find_by_loadbalancer_id(cls, loadbalancer_id, db_session=None):
        return cls.find_by_attribute('loadbalancer_id', loadbalancer_id, db_session)
#     # TODO(dougwig) -- later - should add state enum here

#     # def get_lbaas_root(self):
#     #     # TODO(dougwig), later - if vip, lookup and return vip, if lb, same
#     #     raise Foobar()



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import acos_client.errors as acos_errors

import a10_neutron_lbaas.a10_exceptions as a10_ex

LOG = logging.getLogger(__name__)


class PersistHandler(object):

    def __init__(self, c, context, pool, old_pool=None, deprecated_arg=None):
        self.c = c
        self.context = context
        self.pool = pool
        self.c_pers = None
        self.s_pers = None
        # self.old_pool = old_pool

        self.sp_obj_dict = {
            'HTTP_COOKIE': "cookie_persistence",
            'APP_COOKIE': "cookie_persistence",
            'SOURCE_IP': "src_ip_persistence",
        }

        if pool:
            self.name = pool.id

        if pool and pool.sessionpersistence:
            self.sp = pool.sessionpersistence
            if self.sp.type == 'HTTP_COOKIE' or self.sp.type == 'APP_COOKIE':
                self.c_pers = self.name
            elif self.sp.type == 'SOURCE_IP':
                self.s_pers = self.name
            else:
                raise a10_ex.UnsupportedFeature()
        else:
            self.sp = None

    def c_persistence(self):
        return self.c_pers

    def s_persistence(self):
        return self.s_pers

    def create(self):
        if self.sp is None:
            return
        sp_type = self.sp.type
        if sp_type is not None and sp_type in self.sp_obj_dict:
            try:

                m = getattr(self.c.client.slb.template, self.sp_obj_dict[sp_type])
                m.create(self.name)
            except acos_errors.Exists:
                pass

    def delete(self):
        if self.sp is None:
            return

        sp_type = self.sp.type
        if sp_type in self.sp_obj_dict.keys():
            try:
                m = getattr(self.c.client.slb.template, self.sp_obj_dict[sp_type])
                m.delete(self.name)
            except acos_errors.NotExists:
                pass



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import a10_neutron_lbaas.a10_exceptions as a10_ex
from a10_neutron_lbaas.acos import openstack_mappings

import acos_client.errors as acos_errors
import handler_base_v1
import v1_context as a10


LOG = logging.getLogger(__name__)


class VipHandler(handler_base_v1.HandlerBaseV1):

    def vport_meta(self, vip):
        """Get the vport meta, no matter which name was used"""
        vport_meta = self.meta(vip, 'vport', None)
        if vport_meta is None:
            vport_meta = self.meta(vip, 'port', {})
        return vport_meta

    def create(self, context, vip):
        with a10.A10WriteStatusContext(self, context, vip) as c:
            status = c.client.slb.UP
            if not vip['admin_state_up']:
                status = c.client.slb.DOWN

            pool_name = self._pool_name(context, vip['pool_id'])

            p = PersistHandler(c, context, vip, self._meta_name(vip))
            p.create()

            templates = self.meta(vip, "template", {})

            if 'client_ssl' in templates:
                args = {'client_ssl_template': templates['client_ssl']}
                try:
                    c.client.slb.template.client_ssl.create(
                        '', '', '',
                        axapi_args=args)
                except acos_errors.Exists:
                    pass

            if 'server_ssl' in templates:
                args = {'server_ssl_template': templates['server_ssl']}
                try:
                    c.client.slb.template.server_ssl.create(
                        '', '', '',
                        axapi_args=args)
                except acos_errors.Exists:
                    pass

            vport_list = None
            try:
                vip_meta = self.meta(vip, 'virtual_server', {})
                vport_list = vip_meta.pop('vport_list', None)

                c.client.slb.virtual_server.create(
                    self._meta_name(vip),
                    vip['address'],
                    status,
                    vrid=c.device_cfg.get('default_virtual_server_vrid'),
                    axapi_body=vip_meta)
            except acos_errors.Exists:
                pass

            LOG.debug("VPORT_LIST = %s", vport_list)
            if vport_list is None:
                vport_list = [self.vport_meta(vip)]
            for vport, i in zip(vport_list, range(len(vport_list))):
                try:
                    vport_name = str(i) if i else ''

                    c.client.slb.virtual_server.vport.create(
                        self._meta_name(vip),
                        self._meta_name(vip) + '_VPORT' + vport_name,
                        protocol=openstack_mappings.vip_protocols(c, vip['protocol']),
                        port=vip['protocol_port'],
                        service_group_name=pool_name,
                        s_pers_name=p.s_persistence(),
                        c_pers_name=p.c_persistence(),
                        status=status,
                        autosnat=c.device_cfg.get('autosnat'),
                        ipinip=c.device_cfg.get('ipinip'),
                        axapi_body=vport)
                except acos_errors.Exists:
                    pass

            self.hooks.after_vip_create(c, context, vip)

    def update(self, context, old_vip, vip):
        with a10.A10WriteStatusContext(self, context, vip) as c:
            status = c.client.slb.UP
            if not vip['admin_state_up']:
                status = c.client.slb.DOWN

            pool_name = self._pool_name(context, vip['pool_id'])

            p = PersistHandler(c, context, vip, self._meta_name(vip), old_vip)
            p.create()

            templates = self.meta(vip, "template", {})

            if 'client_ssl' in templates:
                args = {'client_ssl_template': templates['client_ssl']}
                c.client.slb.template.client_ssl.update(
                    '', '', '',
                    axapi_args=args)

            if 'server_ssl' in templates:
                args = {'server_ssl_template': templates['server_ssl']}
                c.client.slb.template.server_ssl.update(
                    '', '', '',
                    axapi_args=args)

            vport_meta = self.vport_meta(vip)

            c.client.slb.virtual_server.vport.update(
                self._meta_name(vip),
                self._meta_name(vip) + '_VPORT',
                protocol=openstack_mappings.vip_protocols(c, vip['protocol']),
                port=vip['protocol_port'],
                service_group_name=pool_name,
                s_pers_name=p.s_persistence(),
                c_pers_name=p.c_persistence(),
                status=status,
                autosnat=c.device_cfg.get('autosnat'),
                ipinip=c.device_cfg.get('ipinip'),
                axapi_body=vport_meta)

            self.hooks.after_vip_update(c, context, vip)

    def _delete(self, c, context, vip):
        try:
            c.client.slb.virtual_server.delete(self._meta_name(vip))
        except acos_errors.NotFound:
            pass

        PersistHandler(c, context, vip, self._meta_name(vip)).delete()

    def delete(self, context, vip):
        with a10.A10DeleteContext(self, context, vip) as c:
            self._delete(c, context, vip)
            self.hooks.after_vip_delete(c, context, vip)


class PersistHandler(object):

    def __init__(self, c, context, vip, vip_name, old_vip=None):
        self.c = c
        self.context = context
        self.vip = vip
        self.c_pers = None
        self.s_pers = None
        self.name = vip_name
        self.old_vip = old_vip

        self.sp_obj_dict = {
            'HTTP_COOKIE': "cookie_persistence",
            'SOURCE_IP': "src_ip_persistence",
        }

        self.sp = vip.get("session_persistence")

        if self.sp is not None:
            if self.sp['type'] == 'HTTP_COOKIE':
                self.c_pers = self.name
            elif self.sp['type'] == 'SOURCE_IP':
                self.s_pers = self.name
            else:
                raise a10_ex.UnsupportedFeature()

    def c_persistence(self):
        return self.c_pers

    def s_persistence(self):
        return self.s_pers

    def create(self):
        if self.old_vip is not None:
            vip_sp = self.old_vip.get("session_persistence")
            if vip_sp is not None:

                try:
                    vip_sp_type = vip_sp.get("type")
                    m = getattr(self.c.client.slb.template, self.sp_obj_dict[vip_sp_type])
                    m.delete(self.old_vip.get("id"))
                except acos_errors.NotExists:
                    pass

        if self.sp is not None:
            sp_type = self.sp.get("type")
            if sp_type is not None and sp_type in self.sp_obj_dict:
                try:
                    m = getattr(self.c.client.slb.template, self.sp_obj_dict[sp_type])
                    m.create(self.name)
                except acos_errors.Exists:
                    pass

    def delete(self):

        if self.sp is None:
            return

        sp_type = self.sp.get("type")
        if sp_type in self.sp_obj_dict.keys():
            try:
                m = getattr(self.c.client.slb.template, self.sp_obj_dict[sp_type])
                m.delete(self.name)

            except Exception as e:
                LOG.exception(e)



# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.from neutron.db import model_base

from a10_neutron_lbaas.tests.db.neutron_ext.db import test_a10_device_instance

import a10_neutron_lbaas.neutron_ext.common.constants as constants
import a10_neutron_lbaas.neutron_ext.services.a10_device_instance.plugin as plugin


class TestPlugin(test_a10_device_instance.TestA10DeviceInstanceDbMixin):

    def setUp(self):
        super(TestPlugin, self).setUp()
        self.plugin = plugin.A10DeviceInstancePlugin()

    def test_supported_extension_aliases(self):
        sea = self.plugin.supported_extension_aliases
        self.assertEqual([constants.A10_DEVICE_INSTANCE_EXT], sea)



#!/usr/bin/env python
# flake8: noqa

from setuptools import setup, find_packages

setup(
    name = "a10-neutron-lbaas",
    version = "1.5.1",
    packages = find_packages(),

    author = "A10 Networks",
    author_email = "dougw@a10networks.com",
    description = "A10 Networks Openstack LBaaS Driver Middleware",
    license = "Apache",
    keywords = "a10 axapi acos adc slb load balancer openstack neutron lbaas",
    url = "https://github.com/a10networks/a10-neutron-lbaas",

    long_description = open('README.md').read(),

    classifiers = [
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet',
    ],

    package_data={'': [
        'a10_neutron_lbaas/db/migration/alembic.ini'
    ]},
    include_package_data=True,

    scripts=['scripts/a10-manage', 'scripts/a10-neutron-lbaas-db-manage'],

    install_requires = ['acos-client>=1.3.5']
)



# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from a10_neutron_lbaas.tests.db import session
# from a10_neutron_lbaas.tests import test_case
from a10_neutron_lbaas.tests.unit import test_base


class UnitTestBase(test_base.UnitTestBase):

    def setUp(self):
        super(UnitTestBase, self).setUp()
        (open_session, close_session) = session.fake_session()
        self.open_session = open_session
        self.close_session = close_session

    def tearDown(self):
        self.close_session()



# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy.orm import sessionmaker

from a10_neutron_lbaas.db import api as db_api

Base = db_api.get_base()


def fake_connection(tables=None):
    # Don't pool connections, use a clean memory database each time
    engine = db_api.get_engine('sqlite://')
    # Reuse a single connection so that the created tables exist in the session
    connection = engine.connect()
    Base.metadata.create_all(connection, tables=tables)
    return connection


def fake_session(tables=None):
    connection = fake_connection(tables)
    Session = sessionmaker(bind=connection)

    def make_session():
        session = Session()
        # Turn off enforcing foreign key constraints
        session.execute('PRAGMA foreign_keys=OFF')
        return session

    return (make_session, connection.close)


def a10_neutron_lbaas_models():
    return [model
            for model in Base._decl_class_registry.values()
            if model.__module__.startswith('a10_neutron_lbaas.')]


def fake_migration_connection():
    a10_neutron_lbaas_tables = [model.__tablename__ for model in a10_neutron_lbaas_models()]
    other_tables = [table
                    for table in Base.metadata.sorted_tables
                    if table.name not in a10_neutron_lbaas_tables]
    return fake_connection(other_tables)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""a10_appliance_scaling_group_table

Revision ID: 2782f2bd94aa
Revises: 2a280aba7701
Create Date: 2016-05-19 18:04:17.611550

"""

# revision identifiers, used by Alembic.
revision = '2782f2bd94aa'
down_revision = '2a280aba7701'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        "a10_scaling_policies",
        sa.Column('id',
                  sa.String(36),
                  primary_key=True,
                  nullable=False),
        sa.Column('tenant_id', sa.String(255), nullable=True),
        sa.Column('created_at', sa.DateTime, nullable=False),
        sa.Column('updated_at', sa.DateTime, nullable=False),
        sa.Column('name', sa.String(255), nullable=True),
        sa.Column('description', sa.String(255), nullable=True),

        sa.Column('cooldown', sa.Integer, nullable=False),
        sa.Column('min_instances', sa.Integer, nullable=False),
        sa.Column('max_instances', sa.Integer, nullable=True)
    )

    op.create_table(
        'a10_scaling_alarms',
        sa.Column('id',
                  sa.String(36),
                  primary_key=True,
                  nullable=False),
        sa.Column('tenant_id', sa.String(255), nullable=True),
        sa.Column('created_at', sa.DateTime, nullable=False),
        sa.Column('updated_at', sa.DateTime, nullable=False),
        sa.Column('name', sa.String(255), nullable=True),
        sa.Column('description', sa.String(255), nullable=True),

        sa.Column('aggregation', sa.String(50), nullable=False),
        sa.Column('measurement', sa.String(50), nullable=False),
        sa.Column('operator', sa.String(50), nullable=False),
        sa.Column('threshold', sa.Float(), nullable=False),
        sa.Column('unit', sa.String(50), nullable=False),
        sa.Column('period', sa.Integer, nullable=False),
        sa.Column('period_unit', sa.String(50), nullable=False)
    )

    op.create_table(
        'a10_scaling_actions',
        sa.Column('id',
                  sa.String(36),
                  primary_key=True,
                  nullable=False),
        sa.Column('tenant_id', sa.String(255), nullable=True),
        sa.Column('created_at', sa.DateTime, nullable=False),
        sa.Column('updated_at', sa.DateTime, nullable=False),
        sa.Column('name', sa.String(255), nullable=True),
        sa.Column('description', sa.String(255), nullable=True),

        sa.Column('action', sa.String(50), nullable=False),
        sa.Column('amount', sa.Integer)
    )

    op.create_table(
        'a10_scaling_policy_reactions',
        sa.Column('id',
                  sa.String(36),
                  primary_key=True,
                  nullable=False),
        sa.Column('created_at', sa.DateTime, nullable=False),
        sa.Column('updated_at', sa.DateTime, nullable=False),
        sa.Column('scaling_policy_id',
                  sa.String(36),
                  sa.ForeignKey('a10_scaling_policies.id'),
                  nullable=False),
        sa.Column('position',
                  sa.Integer,
                  nullable=False),
        sa.Column('alarm_id',
                  sa.String(36),
                  sa.ForeignKey('a10_scaling_alarms.id'),
                  nullable=False),
        sa.Column('action_id',
                  sa.String(36),
                  sa.ForeignKey('a10_scaling_actions.id'),
                  nullable=False)
    )

    op.create_table(
        "a10_scaling_groups",
        sa.Column('id', sa.String(36),
                  primary_key=True,
                  nullable=False),

        sa.Column('tenant_id', sa.String(255), nullable=True),
        sa.Column('created_at', sa.DateTime, nullable=False),
        sa.Column('updated_at', sa.DateTime, nullable=False),
        sa.Column('name', sa.String(255), nullable=True),
        sa.Column('description', sa.String(255), nullable=True),

        sa.Column('scaling_policy_id',
                  sa.String(36),
                  sa.ForeignKey('a10_scaling_policies.id'),
                  nullable=True)
    )

    op.create_table(
        "a10_scaling_group_members",
        sa.Column('id', sa.String(36),
                  primary_key=True,
                  nullable=False),
        sa.Column('created_at', sa.DateTime, nullable=False),
        sa.Column('updated_at', sa.DateTime, nullable=False),
        sa.Column('type', sa.String(50), nullable=False),
        sa.Column('scaling_group_id', sa.String(36),
                  sa.ForeignKey('a10_scaling_groups.id'),
                  nullable=False),
        sa.Column('tenant_id', sa.String(255), nullable=True),
        sa.Column('name', sa.String(255), nullable=True),
        sa.Column('description', sa.String(255), nullable=True),
        sa.Column('host', sa.String(255), nullable=False),
        sa.Column('api_version', sa.String(12), nullable=False),
        sa.Column('username', sa.String(255), nullable=False),
        sa.Column('password', sa.String(255), nullable=False),
        sa.Column('protocol', sa.String(255), nullable=False),
        sa.Column('port', sa.Integer, nullable=False),
        sa.Column('nova_instance_id', sa.String(36), nullable=False)
    )

    op.create_table(
        "a10_scaling_group_switches",
        sa.Column('id', sa.String(36),
                  sa.ForeignKey(u'a10_scaling_group_members.id'),
                  primary_key=True,
                  nullable=False),
    )

    op.create_table(
        "a10_scaling_group_workers",
        sa.Column('id', sa.String(36),
                  sa.ForeignKey(u'a10_scaling_group_members.id'),
                  primary_key=True,
                  nullable=False),
    )


def downgrade():
    op.drop_table('a10_scaling_group_workers')
    op.drop_table('a10_scaling_group_switches')
    op.drop_table('a10_scaling_group_members')
    op.drop_table('a10_scaling_groups')
    op.drop_table('a10_scaling_policy_reactions')
    op.drop_table('a10_scaling_alarms')
    op.drop_table('a10_scaling_actions')
    op.drop_table("a10_scaling_policies")



# Copyright 2015, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class CertManagerWrapper(object):
    def __init__(self, certmgr=None, handler=None):
        if certmgr is not None:
            self.certmgr = certmgr
        else:
            self.certmgr = handler.neutron.bcm_factory()

    def get_certificate(self, container_id, **kwargs):
        return self.certmgr.get_cert(container_id, **kwargs)

    def store_cert(self, certificate, private_key, intermediates=None,
                   private_key_passphrase=None, expiration=None,
                   name='A10-LBaaS TLS Cert', **kwargs):
        return self.certmgr.store_cert(certificate, private_key, intermediates=intermediates,
                                       private_key_passphrase=private_key_passphrase,
                                       expiration=expiration, name=name, kwargs=kwargs)

    def delete_cert(self, cert_ref, service_name='A10-LBaaS', resource_ref=None,
                    **kwargs):
        return self.certmgr.delete_cert(cert_ref, service_name=service_name,
                                        resource_ref=resource_ref, kwargs=kwargs)



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import acos_client.errors as acos_errors

from a10_neutron_lbaas.acos import openstack_mappings
from a10_neutron_lbaas import constants
import a10_neutron_lbaas.v2.wrapper_certmgr as certwrapper
import handler_base_v2
import handler_persist
import v2_context as a10


LOG = logging.getLogger(__name__)


class ListenerHandler(handler_base_v2.HandlerBaseV2):
    def __init__(self, a10_driver, openstack_manager, neutron=None, barbican_client=None):
        super(ListenerHandler, self).__init__(a10_driver, openstack_manager, neutron)
        self.barbican_client = barbican_client

    def _set(self, set_method, c, context, listener):
        if self.barbican_client is None:
            self.barbican_client = certwrapper.CertManagerWrapper(handler=self)

        status = c.client.slb.UP
        if not listener.admin_state_up:
            status = c.client.slb.DOWN

        templates = self.meta(listener, "template", {})

        server_args = {}
        cert_data = dict()

        if listener.protocol and listener.protocol == constants.PROTOCOL_TERMINATED_HTTPS:
            if self._set_terminated_https_values(listener, c, cert_data):
                templates["client_ssl"] = {}
                template_name = str(cert_data.get('template_name', ''))
                key_passphrase = str(cert_data.get('cert_pass', ''))
                cert_filename = str(cert_data.get('cert_filename', ''))
                key_filename = str(cert_data.get('key_filename', ''))
            else:
                LOG.error("Could not created terminated HTTPS endpoint.")

        if 'client_ssl' in templates:
            try:
                c.client.slb.template.client_ssl.create(
                    template_name,
                    cert=cert_filename,
                    key=key_filename)
            except acos_errors.Exists:
                c.client.slb.template.client_ssl.update(template_name, cert=cert_filename,
                                                        key=key_filename, passphrase=key_passphrase)

        if 'server_ssl' in templates:
            server_args = {'server_ssl_template': templates['server_ssl']}
            try:
                c.client.slb.template.server_ssl.create(
                    template_name,
                    cert_filename,
                    key_filename,
                    axapi_args=server_args)
            except acos_errors.Exists:
                c.client.slb.template.server_ssl.update(template_name,
                                                        cert_filename,
                                                        key_filename,
                                                        axapi_args=server_args)

        try:
            pool_name = self._pool_name(context, pool_id=listener.default_pool_id)
        except Exception:
            pool_name = None

        persistence = handler_persist.PersistHandler(
            c, context, listener.default_pool)

        vport_meta = self.meta(listener.loadbalancer, 'vip_port', {})

        try:
            set_method(
                self.a10_driver.loadbalancer._name(listener.loadbalancer),
                self._meta_name(listener),
                protocol=openstack_mappings.vip_protocols(c, listener.protocol),
                port=listener.protocol_port,
                service_group_name=pool_name,
                s_pers_name=persistence.s_persistence(),
                c_pers_name=persistence.c_persistence(),
                status=status,
                autosnat=c.device_cfg.get('autosnat'),
                ipinip=c.device_cfg.get('ipinip'),
                axapi_body=vport_meta)
        except acos_errors.Exists:
            pass

    def _set_terminated_https_values(self, listener, c, cert_data):
        is_success = False
        c_id = listener.default_tls_container_id if listener.default_tls_container_id else None

        if c_id:
            try:
                container = self.barbican_client.get_certificate(c_id, check_only=True)
            except Exception as ex:
                container = None
                LOG.error("Exception encountered retrieving TLS Container %s" % c_id)
                LOG.exception(ex)

            if container:
                base_name = (container._cert_container.name if container._cert_container is not None
                             else "")

                cert_data["cert_content"] = container.get_certificate()
                cert_data["key_content"] = container.get_private_key()
                cert_data["cert_pass"] = container.get_private_key_passphrase()

                cert_data["template_name"] = listener.id

                cert_data["cert_filename"] = "{0}cert.pem".format(base_name)
                cert_data["key_filename"] = "{0}key.pem".format(base_name)

                self._acos_create_or_update(c.client.file.ssl_cert,
                                            file=cert_data["cert_filename"],
                                            cert=cert_data["cert_content"],
                                            size=len(cert_data["cert_content"]),
                                            action="import", certificate_type="pem")

                self._acos_create_or_update(c.client.file.ssl_key,
                                            file=cert_data["key_filename"],
                                            cert=cert_data["key_content"],
                                            size=len(cert_data["key_content"]),
                                            action="import")
                is_success = True
        else:
            LOG.error("default_tls_container_id unspecified for listener. Cannot create listener.")
        return is_success

    def _acos_create_or_update(self, acos_obj, **kwargs):
        if acos_obj.exists(kwargs["file"]):
            acos_obj.update(**kwargs)
        else:
            acos_obj.create(**kwargs)

    def _create(self, c, context, listener):
        self._set(c.client.slb.virtual_server.vport.create,
                  c, context, listener)

    def create(self, context, listener):
        with a10.A10WriteStatusContext(self, context, listener) as c:
            self._create(c, context, listener)

    def _update(self, c, context, listener):
        self._set(c.client.slb.virtual_server.vport.update, c, context, listener)

    def update(self, context, old_listener, listener):
        with a10.A10WriteStatusContext(self, context, listener) as c:
            self._update(c, context, listener)

    def _delete(self, c, context, listener):
        try:
            c.client.slb.virtual_server.vport.delete(
                self.a10_driver.loadbalancer._name(listener.loadbalancer),
                self._meta_name(listener),
                protocol=openstack_mappings.vip_protocols(c, listener.protocol),
                port=listener.protocol_port)
        except acos_errors.NotFound:
            pass

    def delete(self, context, listener):
        with a10.A10DeleteContext(self, context, listener) as c:
            self._delete(c, context, listener)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from a10_neutron_lbaas import a10_exceptions as ex
from a10_neutron_lbaas.db import models

import vthunder_per_tenant

LOG = logging.getLogger(__name__)


# This next set of plumbing hooks needs to be used when the vthunder
# scheduler is active, for one vthunder per VIP.
# LBaaS v2 ONLY.

class VThunderPerVIPPlumbingHooks(vthunder_per_tenant.VThunderPerTenantPlumbingHooks):

    def select_device_with_lbaas_obj(self, tenant_id, a10_context, lbaas_obj,
                                     db_session=None, **kwargs):

        if not self.driver.config.get('use_database'):
            raise ex.RequiresDatabase('vThunder orchestration requires use_database=True')

        # If we already have a vThunder, use it.
        # one vthunder per VIP

        missing_instance = (
            'A10 instance mapped to tenant %s is not present in db; '
            'add it back to config or migrate loadbalancers' % tenant_id
        )

        root_id = lbaas_obj.root_loadbalancer.id
        slb = models.A10SLB.find_by(loadbalancer_id=root_id, db_session=db_session)
        if slb is not None:
            d = self.driver.config.get_device(slb.device_name, db_session=db_session)
            if d is None:
                LOG.error(missing_instance)
                raise ex.InstanceMissing(missing_instance)

            LOG.debug("select_device, returning cached instance %s", d)
            return d

        # No? Then we need to create one.

        if kwargs.get('action') != 'create':
            LOG.error(missing_instance)
            raise ex.InstanceMissing(missing_instance)

        device_config = self._create_instance(tenant_id, a10_context, lbaas_obj, db_session)
        self._wait_for_instance(device_config)

        # Now make sure that we remember where it is.

        models.A10SLB.create_and_save(
            tenant_id=tenant_id,
            device_name=device_config['name'],
            loadbalancer_id=root_id,
            db_session=db_session)

        LOG.debug("select_device, returning new instance %s", device_config)
        return device_config



# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from a10_neutron_lbaas.db import models

import test_base

dt = datetime.datetime.fromtimestamp(1458346727.132739)


class TestTenantBindings(test_base.UnitTestBase):

    def test_model_create(self):
        db = self.open_session()
        x = models.A10TenantBinding(
            tenant_id='xxx',
            device_name='yyy',
            created_at=dt)
        db.add(x)
        db.commit()

        db = self.open_session()
        z = list(db.query(models.A10TenantBinding))
        self.assertEqual(len(z), 1)
        self.assertEqual(z[0].tenant_id, 'xxx')
        self.assertEqual(z[0].device_name, 'yyy')
        self.assertTrue(len(z[0].id) == 36)
        self.assertEqual(z[0].created_at, dt)

        db = self.open_session()
        x = models.A10TenantBinding.create_and_save(
            tenant_id='xxx2',
            device_name='yyy2',
            created_at=dt,
            db_session=db)

        db = self.open_session()
        t = models.A10TenantBinding.find_by_tenant_id('xxx2', db_session=db)
        self.assertEqual(t.device_name, 'yyy2')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

try:
    from neutron.db import l3_db
except ImportError:
    pass
try:
    import neutron_lbaas.common.cert_manager.barbican_cert_manager as bcm
    from neutron_lbaas.db.loadbalancer import models as lb_db
except ImportError:
    # v2 does not exist before Kilo
    pass


class NeutronOpsV2(object):

    def __init__(self, handler):
        self.openstack_driver = handler.openstack_driver
        self.plugin = self.openstack_driver.plugin

    def member_get_ip(self, context, member, use_float=False):
        ip_address = member.address
        if use_float:
            fip_qry = context.session.query(l3_db.FloatingIP)
            if (fip_qry.filter_by(fixed_ip_address=ip_address).count() > 0):
                float_address = fip_qry.filter_by(
                    fixed_ip_address=ip_address).first()
                ip_address = str(float_address.floating_ip_address)
        return ip_address

    def member_count(self, context, member):
        return context.session.query(lb_db.MemberV2).filter_by(
            tenant_id=member.tenant_id,
            address=member.address).count()

    def loadbalancer_total(self, context, tenant_id):
        return context.session.query(lb_db.LoadBalancer).filter_by(
            tenant_id=tenant_id).count()

    def pool_get(self, context, pool_id):
        return self.plugin.db.get_pool(context, pool_id)

    def bcm_factory(self):
        return bcm.CertManager()

    def vip_get(self, context, vip_id):
        return self.plugin.db.get_vip(context, vip_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
# flake8: noqa

# You should really import the models you want directly; this is a placeholder
# to allow existing code to run, while it's converted.

from a10_device_instance import A10DeviceInstance
from a10_slb import A10SLB
from a10_tenant_binding import A10TenantBinding
